import * as THREE from '../build/three.module.js';
import Stats from './jsm/libs/stats.module.js';
import { GUI } from './jsm/libs/dat.gui.module.js';
import { OrbitControls } from "./jsm/controls/OrbitControls.js";

class App {
    constructor() {
        this._initialize();
    }

    _initialize() {
        const domWebGL = document.createElement('div');
        document.body.appendChild(domWebGL);
        this._domWebGL = domWebGL;

        this._amount = 10;
        this._count = Math.pow(this._amount, 3);

        const renderer = new THREE.WebGLRenderer({ antialias: true });
        renderer.outputEncoding = THREE.sRGBEncoding;
        domWebGL.appendChild(renderer.domElement);
        this._renderer = renderer;

        const scene = new THREE.Scene();
        //scene.background = new THREE.Color(0xffffff);
        this._scene = scene;

        this._setupInteractions();
        this._setupCamera();
        this._setupLight();
        this._setupModel();
        this._setupGUI();
        this._setupControls();
        //this._setupHelper();
        
        window.onresize = this.resize.bind(this);

        this.resize();
    }

    _setupInteractions() {
        const raycaster = new THREE.Raycaster();
        raycaster._mouse = new THREE.Vector2(1, 1);

        document.addEventListener("mousemove", this._onMouseMove.bind(this));

        this._raycaster = raycaster;
    }

    _setupModel() {
        const geometry = new THREE.IcosahedronGeometry(0.5, 3);
        const material = new THREE.MeshPhongMaterial();
        const mesh = new THREE.InstancedMesh(geometry, material, this._count);

        let i = 0;
        const offset = (this._amount - 1) / 2;
        const matrix = new THREE.Matrix4();
        const color = new THREE.Color();

        for(let x=0; x<this._amount; x++) {
            for(let y=0; y<this._amount; y++) {
                for(let z=0; z<this._amount; z++) {
                    matrix.setPosition(offset-x, offset-y, offset-z);
                    mesh.setMatrixAt(i, matrix);
                    mesh.setColorAt(i, color);

                    i++;
                }
            }
        }

        this._mesh = mesh;
        this._scene.add(mesh);
    }

    _setupHelper() {
        //.
    }

    _setupGUI() {
        const gui = new GUI();
        gui.add(this._mesh, "count", 0, this._count);
        
        const stats = new Stats();
        this._domWebGL.appendChild(stats.dom);
        this._stats = stats;
    }

    _setupCamera() {    
        const camera = new THREE.PerspectiveCamera(60, window.innerWidth / window.innerHeight, 1, 100);
        camera.position.set(10, 10, 10);
        camera.lookAt(0, 0, 0);

        this._camera = camera;
    }

    _setupControls() {
        new OrbitControls(this._camera, this._renderer.domElement);
    }

    _setupLight() {
        const light1 = new THREE.HemisphereLight(0xffffff, 0x000088);
        light1.position.set(-1, 1.5, 1);
        this._scene.add(light1);

        const light2 = new THREE.HemisphereLight(0xffffff, 0x880000, 0.5);
        light2.position.set(-1, -1.5, -1);
        this._scene.add(light2);
    }

    _onMouseMove(event) {
        event.preventDefault();

        this._raycaster._mouse.x = (event.clientX / window.innerWidth) * 2 - 1;
        this._raycaster._mouse.y = -(event.clientY / window.innerHeight) * 2 + 1;
    }

    update() {
        this._raycaster.setFromCamera(this._raycaster._mouse, this._camera);
        const intersection = this._raycaster.intersectObject(this._mesh);
        const color = new THREE.Color();

        if(intersection.length > 0) {
            const instanceId = intersection[0].instanceId;
            this._mesh.setColorAt(instanceId, color.setHex(Math.random()*0xffffff));
            this._mesh.instanceColor.needsUpdate = true;
        }
    }

    render() {
        requestAnimationFrame(this.render.bind(this));
        
        this.update();
        this._renderer.render(this._scene, this._camera);
    }

    resize() {
        const aspect = window.innerWidth / window.innerHeight;

        this._camera.aspect = aspect;
        this._camera.updateProjectionMatrix();

        this._renderer.setPixelRatio( window.devicePixelRatio );
        this._renderer.setSize(window.innerWidth, window.innerHeight);
    }
}

window.onload = function () {
    (new App()).render();
}